@@ -32,6 +32,11 @@
 #include <errno.h>
 #include <stdarg.h>
 #include <stdbool.h>
+#include <time.h>
+
+#ifdef HAVE_GETPWUID
+#include <pwd.h> /* For alternative way to find HOME dir */
+#endif
 
 #include "cmd_flags.h"
 #include "plugin_name.h"
@@ -230,6 +235,25 @@ static void ERROR(const char * fmt, ...)
      exit(EXIT_FAILURE);
 }
 
+/**
+ * Get the home directory
+ *
+ * @return a pointer to const string that contains home directory
+ */
+static const char * get_home_dir(void)
+{
+     const char * home = getenv("HOME");
+#ifdef HAVE_GETPWUID
+     if(home == NULL)
+     {
+          struct passwd * pw = getpwuid(getuid());
+          home = pw->pw_dir;
+     }
+#endif
+     return home;
+}
+
+
 /**
  * Check the path exists and if so allocate memory to hold the path name
  *
@@ -270,30 +294,33 @@ static char * find_helper_file(const cfgPath_t * paths, const char * basename)
 
           if(paths->envStr)
           {
-               char * tmp = getenv(paths->envStr);
+               const char * tmp = (strcmp("HOME", paths->envStr) == 0) ? get_home_dir() : getenv(paths->envStr);
                if(tmp)
                {
-                    char * wip = tmp = strdup(tmp);
-                    char * sep = NULL;
-                    do
+                    char * const tmp2 = strdup(tmp);
                     {
-                         if( (sep = strchr(wip, ':')) != NULL)
+                         char * wip = tmp2;
+                         char * sep = NULL;
+                         do
                          {
-                              *sep = '\0';
-                              if(sep[-1] == '/')
+                              if( (sep = strchr(wip, ':')) != NULL)
                               {
-                                   sep[-1] = '\0';
+                                   *sep = '\0';
+                                   if(sep[-1] == '/')
+                                   {
+                                       sep[-1] = '\0';
+                                   }
+                              }
+                              snprintf(fname, sizeof(fname), paths->fmtStr, wip, basename);
+                              retVal = chk_path_and_alloc(fname);
+                              if(sep)
+                              {
+                                    wip = &sep[1];
                               }
                          }
-                         snprintf(fname, sizeof(fname), paths->fmtStr, wip, basename);
-                         retVal = chk_path_and_alloc(fname);
-                         if(sep)
-                         {
-                             wip = &sep[1];
-                         }
+                         while(sep && !retVal);
                     }
-                    while(sep && !retVal);
-                    free(tmp);
+                    free(tmp2);
                }
           }
           else
@@ -319,8 +346,8 @@ static const char * find_application(const char * file)
      /* Places to search for the applications */
      static const cfgPath_t binPaths[] =
      {
-          {"%s/%s",                  "PATH"},
-          {NULL,                     NULL}
+          {"%s/%s", "PATH"},
+          {NULL,    NULL}
      };
 
      struct stat filestat;
@@ -488,6 +515,7 @@ static int parse_flags(char **x, command_t *commandp)
      };
 
      const str2flag_t *f;
+     int retVal = false;
 
      for (f = flags; f->name; f++)
      {
@@ -503,7 +531,8 @@ static int parse_flags(char **x, command_t *commandp)
 	            if(p)
                     {
                          *x = p;
-                         return 1;
+                         retVal = true;
+                         break;
                     }
                }
                else if(f->value & H_FMATCH)
@@ -512,16 +541,18 @@ static int parse_flags(char **x, command_t *commandp)
                     if(p)
                     {
                          *x = p;
-                         return 1;
+                         retVal = true;
+                         break;
                     }
                }
                else
                {
-	            return 1;
+	            retVal = true;
+                    break;
                }
 	  }
      }
-     return 0;
+     return retVal;
 }
 
 /**
@@ -658,7 +689,6 @@ static command_t * parse_cfg_cmd(char * buffer)
 	  else if (!parse_flags(&x, cmd))
 	  {
                ERROR("unknown directive %s\n", x);
-               exit(EXIT_FAILURE);
           }
      }
 
@@ -666,7 +696,6 @@ static command_t * parse_cfg_cmd(char * buffer)
      if (*x != ':')
      {
           ERROR("Config error - no command? (%s)\n", x);
-          exit(EXIT_FAILURE);
      }
 
      x = skip_spaces(&x[1]);
@@ -822,7 +851,6 @@ static handler_t * create_handler(pluginType_t * plugin)
      if(!handler)
      {
           ERROR("malloc error\n");
-          exit(EXIT_FAILURE);
      }
      memset(handler, 0, sizeof(handler_t));
 
@@ -1227,7 +1255,7 @@ static void write_handler(handler_t * handler, FILE * mimetypes_fp, FILE * cmds_
 static char * get_processed_dir(void)
 {
      char path[MAX_FILE_PATH_LEN];
-     const char * home = getenv("HOME");
+     const char * home = get_home_dir();
      const char * alt = getenv("MOZPLUGGER_HOME");
      int lenSoFar;
 
@@ -1497,6 +1525,28 @@ static void write_plugin(const char * config_fname, const char * localPluginDir,
      free(path);
 }
 
+/**
+ * Write a timestamp file to indicate the last time mozplugger-update was
+ * called.
+ */
+static void write_ts_file(void)
+{
+     char * path = get_processed_dir();
+
+     if(path)
+     {
+          char ts_fname[MAX_FILE_PATH_LEN];
+          FILE * fp;
+
+          snprintf(ts_fname, sizeof(ts_fname)-1, "%s/.last_update", path);
+          fp = fopen(ts_fname, "wb");
+          fprintf(fp, "%lu\n", time(NULL));
+          fclose(fp);
+
+          free(path);
+     }
+}
+
 /**
  * main, look for and read the mozpluggerrc and create a cached processed
  * versions of that file.
@@ -1559,10 +1609,13 @@ int main(int argc, char * argv[])
           {"%s/.opera/plugins", operaCfgPaths},
      };
 
-     const char * home = getenv("HOME");
+     const char * home = get_home_dir();
      int i;
      int cfgIndex = 0;
 
+     /* first thing make sure processed directory exists */
+     write_ts_file();
+
      if((argc > 1) && (strcmp(argv[1], "-v") == 0))
      {
           g_verbose = 1;
@@ -1594,6 +1647,8 @@ int main(int argc, char * argv[])
                ERROR("Failed to locate mozpluggerrc\n");
           }
 
+          LOG_INFO("Creating plugins for '%s'\n", localPluginDir);
+
           plugins = trim(do_read_config(config_fname));
 
           for(plugin = plugins; plugin ; plugin = plugin->pNext)
