@@ -35,10 +35,15 @@
 #include <errno.h>
 #include <stdarg.h>
 #include <time.h>
+#include <utime.h>
 
 #include <X11/Xlib.h>
 #include <X11/Xutil.h>
 
+#ifdef HAVE_GETPWUID
+#include <pwd.h> /* For alternative way to find HOME dir */
+#endif
+
 #define XP_UNIX
 #include "npapi.h"
 #include "npruntime.h"
@@ -141,7 +146,7 @@ typedef struct handle
  * Global variables
  */
 
-static char errMsg[512];
+static char errMsg[512] = {0};
 static handler_t * g_handlers = 0;
 
 static const char * g_pluginName = "MozPlugger dummy Plugin";
@@ -252,7 +257,7 @@ static int my_putenv_signed(char *buffer, int bufLen, int offset,
  *
  * @param[in] fmt The format string
  */
-static void reportError(const char * fmt, ...)
+static void reportError(NPP instance, const char * fmt, ...)
 {
      va_list ap;
      va_start(ap, fmt);
@@ -260,10 +265,24 @@ static void reportError(const char * fmt, ...)
      vsnprintf(errMsg, sizeof(errMsg), fmt, ap);
      va_end(ap);
 
+     if(instance)
+     {
+          NPN_Status(instance, errMsg);
+     }
      fprintf(stderr, "%s\n",errMsg);
      D("%s\n", errMsg);
 }
 
+static void clearError(void)
+{
+     errMsg[0] = 0;
+}
+
+static bool haveError(void)
+{
+     return errMsg[0] != 0;
+}
+
 /**
  * Allocate some memory from the static pool. We use a static pool for
  * the database because Mozilla can unload the plugin after use and this
@@ -280,8 +299,7 @@ static void * allocStaticMem(int size)
 
      if(newIdx > MAX_STATIC_MEMORY_POOL)
      {
-          reportError("MozPlugger: config file mozpluggerrc is too big - delete" \
-                                        "some handlers/commands or mimetypes");
+          reportError(NULL, "MozPlugger: config file is too big - delete some handlers/commands or mimetypes");
      }
      else
      {
@@ -315,6 +333,25 @@ static const char * makeStrStatic(const char * str, int len)
      return buf;
 }
 
+/**
+ * Search backwards from end of string buffer until first non-white space
+ * and then terminate string after that point to perform a trim.
+ *
+ * @param[in,out] buf The string to be trimmed.
+ */
+static void trim_trailing_spaces(char * buf)
+{
+     char * end = &buf[strlen(buf)-1];
+     for( ;end >= buf; end--)
+     {
+          if((*end != '\r') && (*end != '\n') && (*end != '\t') && (*end != ' '))
+          {
+               end[1] = '\0';
+               return;
+          }
+     }
+}
+
 /**
  * Trim the line buffer and if the line is empty or contains just a comment
  * return false
@@ -327,19 +364,35 @@ static bool chkCfgLine(char * buffer)
 {
      if(buffer[0] != '#')
      {
-          char * end = &buffer[strlen(buffer)-1];
-          for( ;end >= buffer; end--)
-          {
-               if((*end != '\r') && (*end != '\n') && (*end != '\t') && (*end != ' '))
-               {
-                    end[1] = '\0';
-                    return true;
-               }
-          }
+          trim_trailing_spaces(buffer);
+          return true;
      }
      return false;
 }
 
+static bool is_base_mozplugger(const char * magic)
+{
+     return (magic[0] == '-');
+}
+
+/**
+ * Get the home directory
+ *
+ * @return a pointer to const string that contains home directory
+ */
+static const char * get_home_dir(void)
+{
+     const char * home = getenv("HOME");
+#ifdef HAVE_GETPWUID
+     if(home == NULL)
+     {
+          struct passwd * pw = getpwuid(getuid());
+          home = pw->pw_dir;
+     }
+#endif
+     return home;
+}
+
 /**
  * Get the prefix to the path to the config files based on the magic
  *
@@ -351,33 +404,36 @@ static bool chkCfgLine(char * buffer)
  */
 static int get_cfg_path_prefix(const char * magic, char * buf, int bufLen)
 {
-     int retVal;
-     int prefixLen;
-     const char * home = getenv("HOME");
+     const char * fmt = NULL;
+     int prefixLen = 1;
+     const char * home = get_home_dir();
      const char * alt = getenv("MOZPLUGGER_HOME");
 
-     if(magic[0] == '-')
+     if(is_base_mozplugger(magic))
      {
-         magic = "0:";
-     }
-
-     prefixLen = strchr(magic, ':') - magic;
-
-     if(alt)
-     {
-          retVal = snprintf(buf, bufLen, "%s/.processed/%.*s", alt, prefixLen, magic);
-     }
-     else if(home)
-     {
-          retVal = snprintf(buf, bufLen, "%s/.mozplugger/.processed/%.*s", home, prefixLen, magic);
+          magic = "0";
      }
      else
      {
-          D("HOME not defined\n");
-          buf[0] = '\0';
-          retVal = 0;
+          prefixLen = strchr(magic, ':') - magic;
      }
-     return retVal;
+
+     if(alt)
+     {
+          fmt = "%s/.processed/%.*s";
+          home = alt;
+     }
+     else
+     {
+          fmt = "%s/.mozplugger/.processed/%.*s";
+          if(!home)
+          {
+               reportError(NULL, "Mozplugger cannot determine HOME directory\n");
+               *buf = '\0';
+               return 0;
+          }
+     }
+     return snprintf(buf, bufLen, fmt, home, prefixLen, magic);
 }
 
 /**
@@ -862,7 +918,10 @@ static bool do_read_config(const char * magic)
      }
      else
      {
-          reportError("Mozplugger: Installation error - failed to locate mozpluggerrc");
+          if(!haveError())
+          {
+              reportError(NULL, "Mozplugger error - failed to locate %s", config_fname);
+          }
 	  retVal = false;
      }
 
@@ -1217,7 +1276,7 @@ static command_t * find_command(const data_t * THIS, int streamOnly)
 const char * NP2_GetPluginVersion(const char * magic)
 {
      D("NP_GetPluginVersion(%s)\n", magic);
-     if(magic[0] != '-')
+     if(!is_base_mozplugger(magic))
      {
           get_helper_paths(magic);
      }
@@ -1229,7 +1288,7 @@ const char * NP2_GetPluginVersion(const char * magic)
  *
  * @return true if success.
  */
-static bool mozplugger_update(void)
+static bool mozplugger_update(bool * pDoesntExist)
 {
      bool success = true;
      pid_t pid;
@@ -1244,6 +1303,10 @@ static bool mozplugger_update(void)
      else if(pid == 0)
      {
 	  execlp("mozplugger-update", "mozplugger-update", NULL);
+          if( errno == EEXIST)
+          {
+               exit(1000);
+          }
           exit(EXIT_FAILURE);
      }
      else
@@ -1256,10 +1319,18 @@ static bool mozplugger_update(void)
                D("mozplugger-update dumped core or something...\n");
                success = false;
          }
-         else if(WEXITSTATUS(status) != EXIT_SUCCESS)
+         else 
          {
-               D("mozplugger-update exited with status: %d\n", WEXITSTATUS(status));
-               success = false;
+               status = WEXITSTATUS(status);
+               if(status != EXIT_SUCCESS)
+               {
+                    D("mozplugger-update exited with status: %d\n", status);
+                    success = false;
+                    if(status == 1000)
+                    {
+                         *pDoesntExist = true;
+                    }
+               }
          }
      }
      D("mozplugger-update done\n");
@@ -1270,7 +1341,14 @@ static bool mozplugger_update(void)
      return success;
 }
 
-static bool chkValidLocalPluginDirs(void)
+/**
+ * Check is the local plugin directories exist for various browsers
+ * If they do then its likely that the user uses those browsers
+ * so check mozplugger0.so exists in that directory
+ *
+ * @return false if we need to run update
+ */
+static bool chkValidLocalPluginDirs()
 {
      static const char * browsers[] =
      {
@@ -1279,12 +1357,12 @@ static bool chkValidLocalPluginDirs(void)
           "%s/.opera/plugins"
      };
 
-     const char * home = getenv("HOME");
+     const char * home = get_home_dir();
      int i;
 
      if(home == NULL)
      {
-           D("HOME not defined\n");
+           reportError(NULL, "Mozplugger cannot determine HOME directory");
            return false;
      }
 
@@ -1309,81 +1387,142 @@ static bool chkValidLocalPluginDirs(void)
 }
 
 /**
- * Skip over the characters to end marker
+ * Check the last time we updated the cache
  *
- * @return the line end marker
+ * return -1 too soon, +1 about time to update
  */
-static char * skip_to_end(char * p)
+static time_t chkTimeToUpdate(bool * update, bool * dont_update)
 {
-     while((*p != '\r') && (*p != '\n') && (*p != '\0'))
-     {
-          p++;
-     }
-     return p;
-}
+     struct stat details;
+     char ts_fname[256];
+     time_t ts_ftime = 0;
 
-/**
- * Skip end marker
- *
- * @return the start of next line
- */
-static char * skip_end(char * p)
-{
-     while((*p == '\r') || (*p == '\n'))
+     get_cfg_path_prefix(".last_update:", ts_fname, sizeof(ts_fname));
+     if(stat(ts_fname, &details) == 0)
      {
-          p++;
+          time_t now = time(NULL);
+          ts_ftime = details.st_mtime;
+
+          if(ts_ftime > now)
+          {
+               D("Strange .last_update written in the future? %lu s\n", ts_ftime - now); 
+          }
+          else  
+          {
+               time_t diff = now - ts_ftime;
+               if(diff < 10)
+               {
+                    D("Dont update, too soon %lu s\n", diff); 
+                    *dont_update = true;
+               }
+#ifdef AUTO_UPDATE
+               else if(diff > 7*24*60*60)
+               {
+                    D("Auto update %lu s\n", diff); 
+                    *update = true;
+               }
+          }
+#endif
      }
-     return p;
+     return ts_ftime;
 }
 
 /**
  * Parse buf for text containing the VERSION of the config file. Check the
  * version matches
  *
- * @return NULL if not newer, else pointer to next line in buf
+ * @param[in] buf The string read from the description file
+ *
+ * @return false if not matches
  */
-static char * chk_version_matches(char * buf)
+static bool chk_version_matches(char * buf)
 {
-     char * next = skip_to_end(buf);
-     *next = '\0';
-     next = skip_end(next + 1);
      D("Processed config version = '%s'\n", &buf[1]);
+     trim_trailing_spaces(buf);
      if(strcmp(&buf[1],  VERSION) != 0)
      {
           D("Processed config format mismatch should be" VERSION "\n");
-          next = NULL;
+          return false;
      }
-     return next;
+     return true;
 }
  
 /**
  * Parse buf for text containing the name of the config file. Check the
  * date and time of that file against the time of the cached file.
  *
- * @return NULL if not newer, else pointer to next line in buf
+ * @param[in] buf The string read from the description file
+ *
+ * @return false if description file written before timestamp
  */
-static char * chk_cached_is_newer(char * buf, time_t cached_ftime)
+static bool chk_cached_is_newer(char * buf, time_t cached_ftime)
 {
      char * q = strstr(buf, "autogenerated from ");
      if(q)
      {
           struct stat details;
-          char * next;
-          q += 19;
-          next = skip_to_end(q);
-          *next = '\0';
-          next = skip_end(next + 1);
+          q += 19;                  /* Skip text part */
+          trim_trailing_spaces(q);
 
-          if( (stat(q, &details) == 0) && (details.st_mtime > cached_ftime))
+          if( (stat(q, &details) == 0) && (details.st_mtime <= cached_ftime))
           {
-               buf = NULL;
+               return true;
           }
           else
           {
-               buf = next;
+               D("mozpluggerrc = %s %u - %u\n", q, (unsigned) details.st_mtime, (unsigned) cached_ftime);
           }
     }
-    return buf;
+    return false;
+}
+
+static char * read_desc(const char * fname, time_t ts_ftime, bool *update, bool is_base)
+{
+     char * desc = NULL;
+     FILE * fp = fopen(fname, "rb");
+
+     D("Reading '%s'\n", fname);
+     if(fp != NULL)
+     {
+          char linebuf[256];
+          
+          if( fgets(linebuf, sizeof(linebuf), fp)
+                  && chk_version_matches(linebuf) 
+                      && fgets(linebuf, sizeof(linebuf), fp)
+                          && chk_cached_is_newer(linebuf, ts_ftime))
+          {
+                while( fgets(linebuf, sizeof(linebuf), fp) && (linebuf[0] == '#'))
+                     ;
+               
+                if(!is_base)
+                {
+                     struct stat details;
+
+                     fstat(fileno(fp), &details);
+                     desc = malloc(details.st_size+1);
+                     if(desc)
+                     {
+                          D("Size '%u'\n", (unsigned) details.st_size);
+
+                         strcpy(desc, linebuf);
+                         fgets(&desc[strlen(linebuf)], details.st_size, fp);
+                    }
+               }
+          }
+          else
+          {
+               *update = true;
+          }
+          fclose(fp);
+     }
+     else
+     {
+          D("Failed to read description\n");
+#ifdef AUTO_UPDATE
+          *update = true;
+#endif
+     }
+     return desc;
 }
 
 /**
@@ -1397,95 +1536,57 @@ static char * chk_cached_is_newer(char * buf, time_t cached_ftime)
 const char * NP2_GetMIMEDescription(const char * magic)
 {
      char * fname;
-     char * desc = NULL;
-     struct stat details;
+     char * desc;
      bool update = false;
+     bool dont_update = false;
+     bool doesnt_exist = false;
+     time_t ts_ftime;
 
      D("NP_GetMIMEDescription(%s)\n", magic);
 
      if(!chkValidLocalPluginDirs())
      {
           D("Local plugin dirs not valid");
-          desc = "dummy/dummy:*.dummy:Dummy";
           update = true;
      }
 
+     /* Check the last time we updated the cache */
+     ts_ftime = chkTimeToUpdate( &update, &dont_update);
+
+     if(update && !dont_update)
+     {
+          mozplugger_update(&doesnt_exist);
+          ts_ftime = time(NULL);
+          dont_update = true;
+          update = false;
+     }     
+
      fname = get_mimetypes_cfg_path(magic);
+     desc = read_desc(fname, ts_ftime, &update, is_base_mozplugger(magic));
 
-     D("Reading '%s'\n", fname);
-     if(stat(fname, &details) == 0)
+     if(update && !dont_update)
      {
-          time_t cached_ftime = details.st_mtime;
-          int fd = open(fname, O_RDONLY);
- #ifdef AUTO_UPDATE
-          time_t now = time(NULL);
-          D("Auto update %lu - %lu\n", now, cached_ftime); 
-          if((cached_ftime < now) && (now - cached_ftime > 7*24*60*60))
-          {
-              update = true;
-          }
-#endif
-          if(fd >= 0)
-          {
-               char * buf = malloc(details.st_size + 1);
-               char * p = buf;
-               if(buf)
-               {
-                    int bytes = read(fd, buf, details.st_size);
-                    buf[bytes] = '\0';
+          mozplugger_update(&doesnt_exist);
+          ts_ftime = time(NULL);
+          update = false;
+
+          free(desc);
+          desc = read_desc(fname, ts_ftime, &update, is_base_mozplugger(magic));
 
-                    p = chk_version_matches(buf);
-                    if(!p)
-                    {
-                         update = true;
-                         free(buf);
-                         buf = NULL;
-                    }
-                    else
-                    {
-                         p = chk_cached_is_newer(p, cached_ftime);
-                         if(!p)
-                         {
-                              update = true;
-                              free(buf);
-                              buf = NULL;
-                         }
-                    }
-               }
-               close(fd);
-               if(buf && (magic[0] != '-'))
-               {
-                    while(*p == '#')
-                    {
-                        p = skip_end(skip_to_end(p));
-                    }
-                    desc = p;
-               }
-               else
-               {
-                    free(buf);
-               }
-          }
-     }
-     else
-     {
-          D("Failed to read description\n");
-          desc = "dummy/dummy:*.dummy:Dummy";
-#ifdef AUTO_UPDATE
-          update = true;
-#endif
      }
      free(fname);
 
-     if(update)
+     if(!desc && update && !doesnt_exist && !haveError())
      {
-          if(!mozplugger_update() && !desc)
-          {
-               desc = "dummy/dummy:*.dummy:Dummy";
-          }     
+          reportError(NULL, "Please run mozplugger-update");
      }
 
-     D("Getmimedescription done: %s\n", desc);
+     if(haveError())
+     {
+          desc = realloc(desc, 512);
+          snprintf(desc, 511, "dummy/dummy:*.dummy:%s", errMsg);
+     }
+     D("Getmimedescription done: %.100s ...\n", desc);
      return (const char *)desc;
 }
 
@@ -1520,7 +1621,7 @@ bool is_playing(NPP instance)
  */
 static const char * getPluginName(const char * magic)
 {
-     if(magic[0] != '-')
+     if(!is_base_mozplugger(magic))
      {
           get_helper_paths(magic);
      }
@@ -1541,7 +1642,7 @@ static const char * getPluginDescription(const char * magic)
      char * config_fname = get_cmds_cfg_path(magic);
      struct stat details;
 
-     if((magic[0] == '-') || (!config_fname) || (stat(config_fname, &details) != 0))
+     if(is_base_mozplugger(magic) || (!config_fname) || (stat(config_fname, &details) != 0))
      {
           snprintf(desc_buffer, sizeof(desc_buffer),
 		   "MozPlugger version " VERSION
@@ -2165,13 +2266,13 @@ static void new_child(NPP instance, const char* fname, int isURL)
      if(!safeName(fname, isURL)
              || (THIS->urlFragment && !safeName(THIS->urlFragment, 0)))
      {
-	  NPN_Status(instance, "MozPlugger: Detected unsafe URL aborting!");
+	  reportError(instance, "MozPlugger: Detected unsafe URL aborting!");
 	  return;
      }
 
      if (socketpair(AF_UNIX, SOCK_STREAM, 0, commsPipe) < 0)
      {
-	  NPN_Status(instance, "MozPlugger: Failed to create a pipe!");
+	  reportError(instance, "MozPlugger: Failed to create a pipe!");
 	  return;
      }
 
@@ -2224,7 +2325,7 @@ static void new_child(NPP instance, const char* fname, int isURL)
 
      if(THIS->pid == -1)
      {
-	  NPN_Status(instance, "MozPlugger: Failed to fork helper!");
+	  reportError(instance, "MozPlugger: Failed to fork helper!");
      }
 
      D("Child running with pid=%d\n", THIS->pid);
@@ -2471,8 +2572,7 @@ NPError NPP_NewStream(NPP instance, NPMIMEType type, NPStream *stream,
 
      if(THIS->command == 0)
      {
-	  NPN_Status(instance,
-                         "MozPlugger: No appropriate application found.");
+	  reportError(instance, "MozPlugger: No appropriate application found.");
 	  return NPERR_GENERIC_ERROR;
      }
 
@@ -2484,7 +2584,7 @@ NPError NPP_NewStream(NPP instance, NPMIMEType type, NPStream *stream,
 
           if(THIS->tmpFileFd < 0)
           {
-	       NPN_Status(instance, "MozPlugger: Failed to create tmp file");
+	       reportError(instance, "MozPlugger: Failed to create tmp file");
 	       return NPERR_GENERIC_ERROR;
           }
           else
@@ -2646,15 +2746,14 @@ NPError NPP_SetWindow(NPP instance, NPWindow* window)
                 * these types (mms) of urls */
                if (!(THIS->command = find_command(THIS, 1)))
                {
-                    if(*errMsg)
+                    if(haveError())
                     {
                          NPN_Status(instance, errMsg);
-                         *errMsg = '\0';
+                         clearError();
                     }
                     else
                     {
-	                 NPN_Status(instance,
-                          "MozPlugger: No appropriate application found.");
+	                 reportError(instance, "MozPlugger: No appropriate application found.");
                     }
  	            return NPERR_GENERIC_ERROR;
                }
